import unittest

from langchain_core.messages import AIMessage, HumanMessage
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.tools import tool
from langchain_openai import ChatOpenAI
from pydantic import BaseModel

from yid_langchain_extensions.llm.tools_in_prompt_llm import (
    DeepseekR1JsonToolCallsParser, ModelWithPromptIntroducedTools)


@tool
def add(a, b):
    """Adds a and b"""
    return a + b


class Dot(BaseModel):
    """Multiplies a and b"""
    a: int
    b: int


class TestModelWithTools(unittest.TestCase):
    def setUp(self):
        # actually o4-mini supports tools out of the box, but it is easiest to set up as an example
        test_llm = ChatOpenAI(model_name="o4-mini-2025-04-16", model_kwargs={"service_tier": "flex"})
        self.llm = ModelWithPromptIntroducedTools.wrap_model(base_model=test_llm)

    def test_with_chat_messages(self):
        chain_sequential = self.llm.bind_tools(
            tools=[add, Dot], tool_choice="auto") | DeepseekR1JsonToolCallsParser()

        answer: AIMessage = chain_sequential.invoke([HumanMessage(content="hi")])
        self.assertEqual(len(answer.tool_calls), 0)

    def test_with_prompt_template(self):
        prompt = ChatPromptTemplate.from_messages([HumanMessage(content="{message}")])
        chain_sequential = prompt | self.llm.bind_tools(
            tools=[add, Dot], tool_choice="auto") | DeepseekR1JsonToolCallsParser()
        answer: AIMessage = chain_sequential.invoke(input={"message": "hi"})
        self.assertEqual(len(answer.tool_calls), 0)

    def test_auto_sequential_tool_choice(self):
        chain_sequential = self.llm.bind_tools(
            tools=[add, Dot], tool_choice="auto") | DeepseekR1JsonToolCallsParser()

        answer: AIMessage = chain_sequential.invoke("hi")
        self.assertEqual(len(answer.tool_calls), 0)

        answer: AIMessage = chain_sequential.invoke("call tool add(3,5)")
        self.assertEqual(len(answer.tool_calls), 1)
        self.assertEqual(answer.tool_calls[0]["name"], "add")
        self.assertEqual(answer.tool_calls[0]["args"].get("a"), 3)
        self.assertEqual(answer.tool_calls[0]["args"].get("b"), 5)
        self.assertIsNotNone(answer.tool_calls[0]["id"])

    def test_auto_sequential_tool_choice_tricky(self):
        chain_sequential = self.llm.bind_tools(
            tools=[add, Dot], tool_choice="auto") | DeepseekR1JsonToolCallsParser()

        answer: AIMessage = chain_sequential.invoke("call tool add(3,5) and Dot(2,7)")
        self.assertEqual(len(answer.tool_calls), 1)

    def test_auto_parallel_tool_choice(self):
        chain_parallel = self.llm.bind_tools(
            tools=[add, Dot], tool_choice="auto", parallel_tool_calls=True) | DeepseekR1JsonToolCallsParser()

        answer: AIMessage = chain_parallel.invoke("hi")
        self.assertEqual(len(answer.tool_calls), 0)

        answer: AIMessage = chain_parallel.invoke("call tool add(3,5)")
        self.assertEqual(len(answer.tool_calls), 1)
        self.assertEqual(answer.tool_calls[0]["name"], "add")
        self.assertEqual(answer.tool_calls[0]["args"].get("a"), 3)
        self.assertEqual(answer.tool_calls[0]["args"].get("b"), 5)
        self.assertIsNotNone(answer.tool_calls[0]["id"])

        answer: AIMessage = chain_parallel.invoke("call tool add(3,5) and Dot(2,7)")
        self.assertEqual(len(answer.tool_calls), 2)
        self.assertEqual(answer.tool_calls[0]["name"], "add")
        self.assertEqual(answer.tool_calls[0]["args"].get("a"), 3)
        self.assertEqual(answer.tool_calls[0]["args"].get("b"), 5)
        self.assertIsNotNone(answer.tool_calls[0]["id"])
        self.assertEqual(len(answer.tool_calls), 2)
        self.assertEqual(answer.tool_calls[1]["name"], "Dot")
        self.assertEqual(answer.tool_calls[1]["args"].get("a"), 2)
        self.assertEqual(answer.tool_calls[1]["args"].get("b"), 7)
        self.assertIsNotNone(answer.tool_calls[1]["id"])

    def test_any_sequential_tool_choice(self):
        chain_sequential = self.llm.bind_tools(tools=[add, Dot], tool_choice="any") | DeepseekR1JsonToolCallsParser()

        answer: AIMessage = chain_sequential.invoke("hi")
        self.assertEqual(len(answer.tool_calls), 1)
        self.assertEqual(answer.tool_calls[0]["name"], "add")

        answer: AIMessage = chain_sequential.invoke("call tool add(3,5)")
        self.assertEqual(len(answer.tool_calls), 1)
        self.assertEqual(answer.tool_calls[0]["name"], "add")
        self.assertEqual(answer.tool_calls[0]["args"].get("a"), 3)
        self.assertEqual(answer.tool_calls[0]["args"].get("b"), 5)
        self.assertIsNotNone(answer.tool_calls[0]["id"])

        answer: AIMessage = chain_sequential.invoke("call tool add(3,5) and Dot(2,7)")
        self.assertEqual(len(answer.tool_calls), 1)

    def test_any_parallel_tool_choice(self):
        chain_parallel = self.llm.bind_tools(
            tools=[add, Dot], tool_choice="any", parallel_tool_calls=True) | DeepseekR1JsonToolCallsParser()

        answer: AIMessage = chain_parallel.invoke("call tool add(3,5)")
        self.assertEqual(len(answer.tool_calls), 1)
        self.assertEqual(answer.tool_calls[0]["name"], "add")
        self.assertEqual(answer.tool_calls[0]["args"].get("a"), 3)
        self.assertEqual(answer.tool_calls[0]["args"].get("b"), 5)
        self.assertIsNotNone(answer.tool_calls[0]["id"])

        answer: AIMessage = chain_parallel.invoke("call tool add(3,5) and Dot(2,7)")
        self.assertEqual(len(answer.tool_calls), 2)
        self.assertEqual(answer.tool_calls[0]["name"], "add")
        self.assertEqual(answer.tool_calls[0]["args"].get("a"), 3)
        self.assertEqual(answer.tool_calls[0]["args"].get("b"), 5)
        self.assertIsNotNone(answer.tool_calls[0]["id"])
        self.assertEqual(len(answer.tool_calls), 2)
        self.assertEqual(answer.tool_calls[1]["name"], "Dot")
        self.assertEqual(answer.tool_calls[1]["args"].get("a"), 2)
        self.assertEqual(answer.tool_calls[1]["args"].get("b"), 7)
        self.assertIsNotNone(answer.tool_calls[1]["id"])

    def test_any_parallel_tricky(self):
        chain_parallel = self.llm.bind_tools(
            tools=[add, Dot], tool_choice="any", parallel_tool_calls=True) | DeepseekR1JsonToolCallsParser()

        answer: AIMessage = chain_parallel.invoke("hi")
        self.assertGreater(len(answer.tool_calls), 0)
        for tool_call in answer.tool_calls:
            self.assertIn(tool_call["name"], {"add", "Dot"})

    def test_specific_sequential_tool_choice(self):
        chain_sequential = self.llm.bind_tools(tools=[add, Dot], tool_choice="add") | DeepseekR1JsonToolCallsParser()

        answer: AIMessage = chain_sequential.invoke("hi")
        self.assertEqual(len(answer.tool_calls), 1)
        self.assertEqual(answer.tool_calls[0]["name"], "add")

        answer: AIMessage = chain_sequential.invoke("call tool add(3,5)")
        self.assertEqual(len(answer.tool_calls), 1)
        self.assertEqual(answer.tool_calls[0]["name"], "add")
        self.assertEqual(answer.tool_calls[0]["args"].get("a"), 3)
        self.assertEqual(answer.tool_calls[0]["args"].get("b"), 5)
        self.assertIsNotNone(answer.tool_calls[0]["id"])

        answer: AIMessage = chain_sequential.invoke("call tool add(3,5) and Dot(2,7)")
        self.assertEqual(len(answer.tool_calls), 1)

    def test_specific_parallel_tool_choice(self):
        chain_parallel = self.llm.bind_tools(
            tools=[add, Dot], tool_choice="add", parallel_tool_calls=True) | DeepseekR1JsonToolCallsParser()

        answer: AIMessage = chain_parallel.invoke("call tool add(3,5) and add(2,7)")
        self.assertEqual(len(answer.tool_calls), 2)
        self.assertEqual(answer.tool_calls[0]["name"], "add")
        self.assertEqual(answer.tool_calls[0]["args"].get("a"), 3)
        self.assertEqual(answer.tool_calls[0]["args"].get("b"), 5)
        self.assertIsNotNone(answer.tool_calls[0]["id"])
        self.assertEqual(len(answer.tool_calls), 2)
        self.assertEqual(answer.tool_calls[1]["name"], "add")
        self.assertEqual(answer.tool_calls[1]["args"].get("a"), 2)
        self.assertEqual(answer.tool_calls[1]["args"].get("b"), 7)
        self.assertIsNotNone(answer.tool_calls[1]["id"])

    def test_specific_parallel_tool_choice_tricky(self):
        chain_parallel = self.llm.bind_tools(
            tools=[add, Dot], tool_choice="add", parallel_tool_calls=True) | DeepseekR1JsonToolCallsParser()

        answer: AIMessage = chain_parallel.invoke("hi")
        self.assertGreater(len(answer.tool_calls), 0)
        for tool_call in answer.tool_calls:
            self.assertEqual(tool_call["name"], "add")



import copy
import secrets
import string
import typing
from typing import Sequence, Union, Any, Callable, Optional

from langchain_core.callbacks import CallbackManagerForLLMRun
from langchain_core.exceptions import OutputParserException
from langchain_core.language_models import LanguageModelInput, BaseChatModel
from langchain_core.messages import BaseMessage, SystemMessage, AIMessage, ToolCall, HumanMessage
from langchain_core.output_parsers import JsonOutputParser, BaseCumulativeTransformOutputParser
from langchain_core.outputs import ChatResult
from langchain_core.prompt_values import ChatPromptValue
from langchain_core.runnables import Runnable
from langchain_core.tools import BaseTool
from langchain_core.utils.function_calling import convert_to_openai_tool
from pydantic import Field


def add_tool_calls(base_input: LanguageModelInput, extra_message: str) -> LanguageModelInput:
    if isinstance(base_input, str):
        return [HumanMessage(content=base_input), SystemMessage(content=extra_message)]
    if isinstance(base_input, list):
        return base_input + [SystemMessage(content=extra_message)]
    if isinstance(base_input, ChatPromptValue):
        base_input_copy = copy.deepcopy(base_input)
        base_input_copy.messages.append(SystemMessage(content=extra_message))
        return base_input_copy
    raise NotImplementedError(f"type {type(base_input)} not supported")


def split_thinking_and_output(text: str) -> (str, str):
    start_tag = "<think>"
    end_tag = "</think>"
    start = text.find(start_tag)
    if start == -1:
        return "", text
    end = text.find(end_tag, start)
    if end == -1:
        thoughts = text[start + len(start_tag):]
        return thoughts, ""
    thoughts = text[start + len(start_tag):end]
    output = text[end + len(end_tag):]
    return thoughts.strip(), output.strip()


def generate_call_id(length: int = 24, prefix: str = "call_") -> str:
    alphabet = string.ascii_letters + string.digits
    rand_part = ''.join(secrets.choice(alphabet) for _ in range(length))
    return prefix + rand_part


class DeepseekR1JsonToolCallsParser(BaseCumulativeTransformOutputParser[Any]):
    base_json_parser: JsonOutputParser = Field(default_factory=JsonOutputParser)

    def parse(self, text: str) -> AIMessage:
        thoughts, output = split_thinking_and_output(text)
        try:
            raw_tool_calls = self.base_json_parser.parse(output)
        except OutputParserException as e:
            return AIMessage(content=output, additional_kwargs={"thoughts": thoughts, "parsing_error": str(e)})
        if isinstance(raw_tool_calls, dict):
            raw_tool_calls = [raw_tool_calls]

        tool_calls = []
        try:
            for raw_tool_call in raw_tool_calls:
                tool_calls.append(
                    ToolCall(
                        name=raw_tool_call["name"],
                        args=raw_tool_call["arguments"],
                        id=generate_call_id()
                    )
                )
        except KeyError as e:
            return AIMessage(content=output, additional_kwargs={"thoughts": thoughts, "parsing_error": str(e)})
        return AIMessage(content="", additional_kwargs={"thoughts": thoughts}, tool_calls=tool_calls)


class ModelWithPromptIntroducedTools(BaseChatModel):
    """
    Some models do not support tools calling out-of-the box,
     but they are still smart enough to properly follow the schema.
    So you can use this wrapper for such models, to support binding tools for them.
    Tools will be introduced as part of the input prompt.
    """
    base_model: BaseChatModel

    @classmethod
    def wrap_model(cls, base_model: BaseChatModel) -> "ModelWithPromptIntroducedTools":
        return ModelWithPromptIntroducedTools(
            base_model=base_model,
            name=base_model.name,
            cache=base_model.cache,
            verbose=base_model.verbose,
            callbacks=base_model.callbacks,
            tags=base_model.tags,
            metadata=base_model.metadata,
            custom_get_token_ids=base_model.custom_get_token_ids,
            callback_manager=base_model.callback_manager,
            rate_limiter=base_model.rate_limiter,
            disable_streaming=base_model.disable_streaming,
        )

    def _generate(self, messages: list[BaseMessage], stop: Optional[list[str]] = None,
                  run_manager: Optional[CallbackManagerForLLMRun] = None, **kwargs: Any) -> ChatResult:
        return self.base_model._generate(messages, stop, run_manager, **kwargs)

    @property
    def _llm_type(self) -> str:
        return self.base_model._llm_type

    def bind_tools(
            self,
            tools: Sequence[Union[dict[str, Any], type, Callable, BaseTool]],
            *,
            tool_choice: Optional[
                Union[dict, str, typing.Literal["auto", "none", "required", "any"], bool]
            ] = None,
            strict: Optional[bool] = None,
            parallel_tool_calls: Optional[bool] = None,
            **kwargs: Any,
    ) -> Runnable[LanguageModelInput, BaseMessage]:
        formatted_tools = [
            convert_to_openai_tool(tool, strict=strict) for tool in tools
        ]
        formatted_tools = {
            tool["function"]["name"]: tool for tool in formatted_tools
        }

        match tool_choice:
            case "none":
                return self
            case "any" | "required":
                if parallel_tool_calls:
                    suffix = ("Right now you should call one or several tools from this list. "
                              "It is forbidden to answer with plain text right now. "
                              "Return a list of tool calls (it might be a length of 1)")
                else:
                    suffix = ("Right now you should call one tool from this list. "
                              "It is forbidden to answer with plain text right now. "
                              "Return a tool call json.")
            case "auto" | None | False:
                if parallel_tool_calls:
                    suffix = ("Right now you should call one or several tools from this list. "
                              "Or you can answer with plain text to user instead of calling tools. "
                              "Return a plain text or a list of tool calls (it might be a length of 1)")
                else:
                    suffix = ("Right now you should call some tool from this list. "
                              "Or you can answer with plain text to user instead of calling any tool. "
                              "Return a plain text or a tool call json.")
            case cmd if cmd in formatted_tools.keys():
                formatted_tools = {k: v for k, v in formatted_tools.items() if k == tool_choice}
                suffix = (f"Right now you should call tool '{tool_choice}'. "
                          "Return a tool call json.")
            case _:
                raise ValueError("Unsupported value of tool_choice argument")

        tools_intro = f"You have access to {len(formatted_tools)} tools with following schemas:\n"
        for tool_name, formatted_tool in formatted_tools.items():
            tools_intro += f"{formatted_tool}\n"
        tools_intro += (f"\n{suffix}\nHint: before actually calling the tool,"
                        f" think well, how are you going to call it. "
                        f"During thinking, make sure you precisely follow the tool schema!!! ")

        if parallel_tool_calls:
            tools_intro += """
Example of tools calling:
```json
[
    {
        "name": "tool1_name",
        "arguments": {
            "int_arg_name": int_value,
            "bool_arg_name": true,
        }
    },
    {
        "name": "tool2_name",
        "arguments": {
            "optional_arg_name": null,
            "str_arg_name": "str_value"
        }
    },
]
```
"""
        else:
            tools_intro += """
Example of tool calling:
```json
{
    "name": "tool_name",
    "arguments": {
        "int_arg_name": int_value,
        "bool_arg_name": true,
        "optional_arg_name": null,
        "str_arg_name": "str_value"
    }
}
```
"""

        return (lambda input: add_tool_calls(input, tools_intro)) | self



import copy
from typing import Optional, Any

from langchain_core.exceptions import OutputParserException
from langchain_core.language_models import BaseChatModel, LanguageModelInput, LanguageModelOutput
from langchain_core.messages import AIMessage, ToolMessage, SystemMessage
from langchain_core.prompt_values import ChatPromptValue
from langchain_core.runnables import Runnable, RunnableConfig
from langchain_core.runnables.utils import Input, Output


class LLMWithParsingRetry(Runnable[LanguageModelInput, Any]):
    def __init__(
            self, llm: BaseChatModel, parser: Runnable[LanguageModelOutput, Any],
            max_retries: int = 1, exceptions_to_retry: tuple[type[Exception]] = (OutputParserException,)
    ):
        self.llm = llm
        self.parser = parser
        self.max_retries = max_retries
        self.exceptions_to_retry = exceptions_to_retry

    def _extend_input(
            self, input: LanguageModelInput, bad_result: LanguageModelOutput, error_message: str
    ) -> LanguageModelInput:
        error_message = f"Output parsing error: {error_message}, try again, but follow instructions carefully."
        if isinstance(input, str):
            bad_result_str = bad_result if isinstance(bad_result, str) else bad_result.content
            return input + "\n\n" + f"AI: {bad_result_str}" + "\n\n" + error_message
        bad_result_message = bad_result if isinstance(bad_result, AIMessage) else AIMessage(content=bad_result.content)

        error_messages = []
        if bad_result_message.tool_calls:
            # some llms require all tool calls to have a result:
            for tool_call in bad_result_message.tool_calls:
                error_messages.append(
                    ToolMessage(
                        tool_call_id=tool_call["id"],
                        content="tool call failed because of parsing error (check further)")
                )
        error_messages.append(SystemMessage(content=error_message))
        if isinstance(input, list):
            return input + [bad_result_message] + error_messages
        elif isinstance(input, ChatPromptValue):
            input_copy = copy.deepcopy(input)
            input_copy.messages.append(bad_result_message)
            input_copy.messages.extend(error_messages)
            return input_copy
        raise NotImplementedError(f"type {type(input)} not supported")

    def invoke(self, input: Input, config: Optional[RunnableConfig] = None, **kwargs: Any) -> Any:
        aggregated_error = ""
        extended_input = input
        for _ in range(self.max_retries + 1):
            llm_output = self.llm.invoke(extended_input, config, **kwargs)
            try:
                parser_output = self.parser.invoke(llm_output, config, **kwargs)
                return parser_output
            except self.exceptions_to_retry as e:
                aggregated_error += str(e) + "\n\n"
                extended_input = self._extend_input(extended_input, llm_output, str(e))
        raise OutputParserException(f"Failed to parse LLM output after {self.max_retries} retries:\n{aggregated_error}")

    async def ainvoke(
        self, input: Input, config: Optional[RunnableConfig] = None, **kwargs: Any
    ) -> Output:
        aggregated_error = ""
        extended_input = input
        for _ in range(self.max_retries + 1):
            llm_output = await self.llm.ainvoke(extended_input, config, **kwargs)
            try:
                parser_output = await self.parser.ainvoke(llm_output, config, **kwargs)
                return parser_output
            except self.exceptions_to_retry as e:
                aggregated_error += str(e) + "\n\n"
                extended_input = self._extend_input(extended_input, llm_output, str(e))
        raise OutputParserException(f"Failed to parse LLM output after {self.max_retries} retries:\n{aggregated_error}")



from unittest import TestCase

from langchain_core.messages import AIMessage, HumanMessage, SystemMessage
from langchain_openai import ChatOpenAI

from yid_langchain_extensions.utils import FirstMessageAuthorContextSizeLimiter, \
    NaiveContextSizeLimiter


class TestFirstMessageAuthorContextSizeLimiter(TestCase):
    def test_limit_messages(self):
        messages = [
            HumanMessage(content="hi"),
            SystemMessage(content="say hi"),
            AIMessage(content="hi"),
            HumanMessage(content="how r u?"),
            SystemMessage(content="say how r u"),
            AIMessage(content="im fine"),
        ]
        limiter = FirstMessageAuthorContextSizeLimiter(
            first_message_author="human",
            base_limiter=NaiveContextSizeLimiter(
                max_context_size=40, llm=ChatOpenAI()
            )
        )
        limited_messages = limiter.limit_messages(messages)
        self.assertEqual(len(limited_messages), 3)



import unittest

from langchain_core.messages import AIMessage, HumanMessage
from langchain_core.output_parsers import PydanticOutputParser
from langchain_openai import ChatOpenAI
from pydantic import BaseModel

from yid_langchain_extensions.llm.retrying_llm import LLMWithParsingRetry


class Dot(BaseModel):
    """Multiplies a and b"""
    a: int
    b: int


class TestRetryingLLM(unittest.TestCase):
    def setUp(self):
        self.llm = ChatOpenAI(model_name="gpt-4.1-mini", temperature=0)

    def test_with_chat_messages(self):
        retrying_llm = LLMWithParsingRetry(
            llm=self.llm,
            parser=PydanticOutputParser(pydantic_object=Dot),
            max_retries=3
        )

        answer: AIMessage = retrying_llm.invoke([HumanMessage(content="return json call of function Dot(2,7)")])
        self.assertEqual(answer, Dot(a=2, b=7))


class TestRetryingLLMAsync(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.llm = ChatOpenAI(model_name="gpt-4.1-mini", temperature=0)

    async def test_with_chat_messages(self):
        retrying_llm = LLMWithParsingRetry(
            llm=self.llm,
            parser=PydanticOutputParser(pydantic_object=Dot),
            max_retries=3
        )

        answer: AIMessage = await retrying_llm.ainvoke([HumanMessage(content="return json call of function Dot(2,7)")])
        self.assertEqual(answer, Dot(a=2, b=7))



import unittest

from langchain import hub
from langchain_core.output_parsers import PydanticToolsParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.utils.function_calling import convert_to_openai_tool
from langchain_openai import ChatOpenAI
from pydantic import BaseModel as BaseModelV2, Field as FieldV2

from yid_langchain_extensions.llm.tools_llm_with_thought import build_tools_llm_with_thought


class PowerFnArgsV2(BaseModelV2):
    power: int = FieldV2(description="power")
    base: float = FieldV2(description="base")


class Reasoning(BaseModelV2):
    reasoning: str


class TestToolsLLMWithThought(unittest.TestCase):
    def test(self):
        tools = [convert_to_openai_tool(PowerFnArgsV2)]
        base_prompt = hub.pull("hwchase17/openai-tools-agent")
        thought_prompt: ChatPromptTemplate = hub.pull("dimitree54/introduce_thought_tool")
        llm = ChatOpenAI(model="gpt-4.1-nano", temperature=0)  # noqa
        llm_with_thought = build_tools_llm_with_thought(llm, tools, thought_prompt, Reasoning)
        extended_chain = base_prompt | llm_with_thought | PydanticToolsParser(tools=[PowerFnArgsV2])
        result = extended_chain.invoke({"input": "3.43^5", "agent_scratchpad": [], "chat_history": []})
        self.assertEqual(result[0].base, 3.43)
        self.assertEqual(result[0].power, 5)



import base64
from abc import ABC, abstractmethod
from typing import Dict, Union, Optional, Sequence, List

import cv2
import numpy as np
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_core.prompt_values import ChatPromptValue
from langchain_core.runnables import RunnableConfig, Runnable
from langchain_core.utils.function_calling import _rm_titles, convert_to_openai_function, convert_to_openai_tool  # noqa
from pydantic import BaseModel


class ChatPromptValue2DictAdapter(Runnable[Union[ChatPromptValue, Dict], Dict[str, Sequence[BaseMessage]]]):
    def invoke(
            self, input: Union[ChatPromptValue, Dict], config: Optional[RunnableConfig] = None
    ) -> Dict[str, Sequence[BaseMessage]]:
        if isinstance(input, Dict):
            return input
        return {
            "messages": input.messages
        }


# todo refactor it to be runnable
class ContextSizeLimiter(BaseModel, ABC):
    @abstractmethod
    def limit_messages(self, messages: List[BaseMessage]) -> List[BaseMessage]:
        pass


class NaiveContextSizeLimiter(ContextSizeLimiter):
    max_context_size: int
    llm: BaseChatModel

    def limit_messages(self, messages: List[BaseMessage]) -> List[BaseMessage]:
        num_messages_to_keep = 1
        while num_messages_to_keep <= len(messages):
            extended_num_messages_to_keep = num_messages_to_keep + 1
            last_messages = messages[-extended_num_messages_to_keep:]
            tokens_in_extended_messages = self.llm.get_num_tokens_from_messages(last_messages)
            if tokens_in_extended_messages > self.max_context_size:
                break
            num_messages_to_keep = extended_num_messages_to_keep
        return messages[-num_messages_to_keep:]


class FirstMessageAuthorContextSizeLimiter(ContextSizeLimiter):
    first_message_author: str
    base_limiter: ContextSizeLimiter

    def _remove_head_until_author(self, messages: List[BaseMessage]) -> List[BaseMessage]:
        for i in range(len(messages)):
            if messages[i].type == self.first_message_author:
                return messages[i:]
        return []

    def limit_messages(self, messages: List[BaseMessage]) -> List[BaseMessage]:
        messages = self.base_limiter.limit_messages(messages)
        return self._remove_head_until_author(messages)


def encode_image_to_url(image: np.ndarray) -> str:
    # Ensure the image is in 3-channel format
    if len(image.shape) != 3 or image.shape[2] != 3:
        raise ValueError("Input image must be a 3-channel image")

    # Encode the image to PNG format
    _, buffer = cv2.imencode('.png', image)

    # Convert the buffer to base64
    base64_image = base64.b64encode(buffer).decode('utf-8')

    # Create the data URL
    base64_url = f"data:image/png;base64,{base64_image}"

    return base64_url



import unittest
from typing import Tuple

import numpy as np
from langchain_core.messages import HumanMessage
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI

from yid_langchain_extensions.llm.batches_openai_client import BatchesOpenAICompletions
from yid_langchain_extensions.utils import encode_image_to_url


class TestBatchesClient(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        batches_client = BatchesOpenAICompletions()
        self.llm = ChatOpenAI(model_name="gpt-4.1-nano", temperature=0, async_client=batches_client)

    @staticmethod
    def generate_random_image():
        random_image_array = np.random.randint(0, 256, (128, 128, 3), dtype=np.uint8)
        return random_image_array

    async def test_chain_with_text(self) -> Tuple[bool, str]:
        """Test the batches client with text input in a chain."""
        try:
            prompt = ChatPromptTemplate.from_messages([HumanMessage(content="{message}")])
            chain = prompt | self.llm | StrOutputParser()
            answer = await chain.ainvoke(input={"message": "hi"})
            return len(answer) > 0, "Text chain test passed"
        except Exception as e:
            return False, f"Text chain test failed: {str(e)}"

    async def test_chain_with_inline_image(self) -> Tuple[bool, str]:
        """Test the batches client with an inline image."""
        try:
            chain = self.llm | StrOutputParser()

            content = [
                {'type': 'text', 'text': 'hi'},
                {'type': 'image_url', 'image_url': {
                    'url': encode_image_to_url(self.generate_random_image())
                }}
            ]

            answer = await chain.ainvoke(input=[HumanMessage(content=content)])
            return len(answer) > 0, "Inline image test passed"
        except Exception as e:
            return False, f"Inline image test failed: {str(e)}"

    async def test_chain_with_online_image(self) -> Tuple[bool, str]:
        """Test the batches client with an online image."""
        try:
            chain = self.llm | StrOutputParser()

            url = ("https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/"
                   "Gfp-wisconsin-madison-the-nature-boardwalk.jpg/"
                   "2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg")

            content = [
                {'type': 'text', 'text': 'hi'},
                {'type': 'image_url', 'image_url': {
                    'url': url
                }}
            ]

            answer = await chain.ainvoke(input=[HumanMessage(content=content)])
            return len(answer) > 0, "Online image test passed"
        except Exception as e:
            return False, f"Online image test failed: {str(e)}"



import copy
from typing import Dict, Type, Optional, Any, List, AsyncIterator

from langchain_core.messages import AIMessage, AIMessageChunk
from langchain_core.prompt_values import ChatPromptValue
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable, RunnableConfig, RunnablePassthrough
from langchain_core.utils.function_calling import convert_to_openai_tool
from langchain_openai import ChatOpenAI
from pydantic import BaseModel

from yid_langchain_extensions.utils import ChatPromptValue2DictAdapter


class ThoughtStripper(Runnable[AIMessageChunk, AIMessageChunk]):
    def __init__(self, thought_name: str):
        self.thought_name = thought_name
        self.silenced = False

    def invoke(
            self,
            input: AIMessage,  # noqa
            config: Optional[RunnableConfig] = None
    ) -> AIMessage:
        return self._strip(input)

    async def atransform(
        self,
        input: AsyncIterator[AIMessageChunk],  # noqa
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> AsyncIterator[AIMessageChunk]:
        async for chunk in input:
            yield self._strip_from_chunk(chunk)
        self.silenced = False

    def _strip(self, message: AIMessage) -> AIMessage:
        message = copy.deepcopy(message)
        message.additional_kwargs["tool_calls"] = [tool_call for tool_call in message.additional_kwargs["tool_calls"]
                                                   if tool_call["function"]["name"] != self.thought_name]
        message.tool_calls = [tool_call for tool_call in message.tool_calls
                              if tool_call["name"] != self.thought_name]
        if isinstance(message, AIMessageChunk):
            message.tool_call_chunks = [tool_call for tool_call in message.tool_call_chunks
                                        if tool_call["name"] != self.thought_name]
        return message

    def _strip_from_chunk(self, message: AIMessageChunk) -> AIMessageChunk:
        message = copy.deepcopy(message)
        for tool_call_chunk in message.tool_call_chunks:
            tool_name = tool_call_chunk["name"]
            if tool_name == self.thought_name:
                self.silenced = True
            elif tool_name is not None:
                self.silenced = False
        if self.silenced:
            message.additional_kwargs["tool_calls"] = []
            message.tool_calls = []
            message.tool_call_chunks = []
        return message


def build_tools_llm_with_thought(
        tools_llm: ChatOpenAI,
        openai_tools: List[Dict[str, Any]],
        thought_introducing_prompt: ChatPromptTemplate,
        thought_class: Type[BaseModel]
) -> Runnable[ChatPromptValue, AIMessage]:
    thought_tool = convert_to_openai_tool(thought_class)
    all_tools = [thought_tool] + openai_tools
    llm_with_thought_tool = tools_llm.bind(tools=all_tools)

    thought_tool_name = thought_tool["function"]["name"]
    thought_stripper = ThoughtStripper(thought_name=thought_tool_name)
    return ChatPromptValue2DictAdapter() | RunnablePassthrough.assign(
        thought_tool_name=lambda x: thought_tool_name
    ) | thought_introducing_prompt | llm_with_thought_tool | thought_stripper
